"""
Utilities to help with HTTP posting.
"""
import uuid
import itertools
import os

class MultiPartForm(object):
    """Accumulate the data to be used when posting a form.
    
    Inspired from PMOTW (https://pymotw.com/2/urllib2/).
    
    Ported to Python 3 and specialized to deal with STL file upload to iMaterialise.    
    """
    def __init__(self):
        self.form_fields = []
        self.files = []
        self.boundary = str(uuid.uuid4())
    
    def get_content_type(self):
        return 'multipart/form-data; boundary={}'.format(self.boundary)

    def add_field(self, name, value):
        """Add a simple field to the form data."""
        self.form_fields.append((name, value))    
    
    def add_stl(self, fieldname, filename):
        """Add an STL file to be uploaded."""        
        with open(filename, 'rb') as open_file:
          body = open_file.read()        
        self.files.append((fieldname, os.path.basename(filename), 'application/octet-stream', body))    
    
    def tobytes(self):
        """Return a string representing the form data, including attached files."""
        # Build a list of lists, each containing "lines" of the
        # request.  Each part is separated by a boundary string.
        # Once the list is built, return a string where each
        # line is separated by '\r\n'.  
        parts = []
        part_boundary = b'--' + self.boundary.encode()
        
        # Add the files to upload
        parts.extend(
            [ part_boundary,
              ('Content-Disposition: form-data; name="{}"; filename="{}"'.format(field_name, filename)).encode(),
              ('Content-Type: %s' % content_type).encode(),
              b'',
              body,
            ]
            for field_name, filename, content_type, body in self.files
            )
        
        # Add the form fields
        parts.extend(
            [ part_boundary,
              ('Content-Disposition: form-data; name="%s"' % name).encode(),
              b'',
              value.encode(),
            ]
            for name, value in self.form_fields
            )        
        
        # Flatten the list and add closing boundary marker,
        # then return CR+LF separated data        
        flattened = list(itertools.chain(*parts))
        flattened.append(b'--' + self.boundary.encode() + b'--')
        flattened.append(b'')
        return b'\r\n'.join(flattened) # must be "\r\n", otherwise bad request
