﻿"""
Get information on materials, shipment...

.. autosummary::
  
  get_shipment_info
  materials_catalog
  get_material_info
  get_finish_info
  get_material_and_finish_ID
  
"""
from imaterialise.api import server
if server.PY2:
  import urllib2 as urllib_request
else:
  import urllib.request as urllib_request
import json
import logging
import os

logger = logging.getLogger("API.INFO")

__mat_catalog = None
__mat_catalog_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'mat.cat')


def get_shipment_info(countryCode='BE', stateCode='', city='Leuven', zipCode='3001'):
  """
  Return a dictionary with shipment information.
  
  All shipment information need to be parsed as strings (even if it is a number).
  
  :param countryCode: country code (default: 'BE')
  :type countryCode: str
  :param stateCode: state code (defaults to empty string)
  :type stateCode: str
  :param city: city name (default: Leuven)
  :type city: str
  :param zipCode: city zipcode (default: 3001)
  :type zipCode: str
  :return: shipment info
  :rtype: dict
  """
  return dict(countryCode=countryCode, stateCode=stateCode, city=city, zipCode=str(zipCode))
  

def materials_catalog(force_update=False):
  """
  Get the list of materials and available finishes per material from the server.
  
  See `Materials Catalog API <https://i.materialise.com/api/docs/materials-catalog-api/>`_
  for an example response.
  
  The server is forced to be the genuite i.materialise.com server (i.e. sandbox is not allowed)
  
  Example usage: to retrieve the list of materials and their finishes, execute::
  
    cat = info.materials_catalog()
    tuples  = [(m['name'], f['name']) for m in cat['materials'] for f in m['finishes']]
    for (material, finish) in tuples:
      print(material, finish)
   
  Gives::
    
    Polyamide Natural white
    Polyamide Polished natural white
    ...
    Silver High gloss
    Silver Gloss
    ...      
  
  :return: Dictionary with all materials and their properties
  :rtype: dict
  """
  
  if __mat_catalog is None or force_update:
    server_name = 'i.materialise.com'
    url = 'https://{server_name}/web-api/materials?user={user}'.format(server_name=server_name,
                      user=server.properties['user'])
    logger.debug("Querying materials catalog at {}".format(url))
    response = urllib_request.urlopen(url)
    output = json.loads(response.read().decode('utf-8'))
    response.close()
    
    # Overwrite existing file
    with open(__mat_catalog_file, 'w') as open_cat:
        open_cat.write(json.dumps(output))
    globals()['__mat_catalog'] = output
    
  else:
    output = __mat_catalog
  return output
  

def get_material_info(material_name):
  """
  Return information on one specific material type.
  
  :param material_name: name of the requested material (case insensitive, white space sensitive)
  :type material_name: str
  :return: material info or None when the material is not found
  :rtype: dict or None
  """
  material_name = material_name.lower()
  for mat in __mat_catalog['materials']:
    if mat['name'].lower() == material_name:
      return mat

      
def get_finish_info(material_name, finish_name):
  """
  Return information on one specific material type.
  
  :param material_name: name of the requested material (case insensitive, white space sensitive)
  :type material_name: str
  :param finish_name: name of the requested finish (case insensitive, white space sensitive)
  :type finish_name: str
  :return: finish info or None when the finish is not found for the given material
  :rtype: dict or None
  """
  mat_info = get_material_info(material_name)  
  if mat_info is not None:
    finish_name = finish_name.lower()
    for fin in mat_info['finishes']:
      if fin['name'].lower() == finish_name:
        return fin
 
 
def get_material_and_finish_ID(material_name, finish_name):
  """
  Retrieve the ID of a material and its finish.
  
  Raises a ValueError if any of the material or finish cannot be found.
  """
  # Retrieve material ID
  if material_name is None:
    material_name = __mat_catalog['materials'][0]['name']
  mat_info = get_material_info(material_name)
  if mat_info is not None:
    materialID = mat_info['materialID']
  else:
    raise ValueError("Material `{}` is not available".format(material_name))
  
  # Retrieve finish ID
  if finish_name is None:
    finish_name = mat_info['finishes'][0]['name']
  finish_info = get_finish_info(material_name, finish_name)
  if finish_info is not None:
    finishID = finish_info['finishID']
  else:
    list_available = ", ".join([f['name'] for f in mat_info['finishes']])
    msg = "Finish `{}` is not available for material `{}` (not any of {})"
    raise ValueError(msg.format(finish_name, material_name, list_available))
  
  return materialID, finishID

  
  
def get_demo_stl():
  """
  Return the file path to a demo STL file.
  
  :return: file path to STL demo file
  :rtype: str
  """
  return os.sep.join(os.path.abspath(__file__).split(os.sep)[:-2] + ['files', 'stls', 'Cylinder.stl'])    
      
  
# Always make sure there is a default list of materials,
# even if the server is down
if __mat_catalog is None:
  with open(__mat_catalog_file, 'r') as open_cat:
    __mat_catalog = json.loads(open_cat.read())
  