﻿"""
Ordering APIs

.. autosummary::
  
  upload_model
  print_lab_connection
  
"""
from imaterialise.api import server
from imaterialise.api import httputils
from imaterialise.api import info
import os
import json
import logging
import webbrowser
if server.PY2:
  import urllib2 as urllib_request
  import urllib2.error as urllib_error
else:
  import urllib.request as urllib_request
  import urllib.error as urllib_error

logger = logging.getLogger("API.ORDERING")

def upload_model(stl_file=None, file_units='mm', toolID=None):
  """
  Retrieve the dimensions of a model by uploading it to the iMaterialise server
  
  The return value is a dictionary containing the keys:
    - toolID
    - modelID, modelFilename
    - fileUnits
    - xDimMm, yDimMm, zDimMm
    - volumeCm3, surfaceCm2
    - modelStatus
    - validUntil
    
  The ``modelStatus`` can be any of:
    - succeeded
    - neededFixing
    - stillNeedsFixing
    - backEndException
  
  See `Upload model <https://i.materialise.com/api/docs/upload-model-api/>`_ for more
  information.
  
  Errors:
    - if the STL files does not exist, an IOError is raised
    - if the units are invalid, a ValueError is raised
    - if the server cannot be contacted, an URLOpenError is raised
  
  :param stl_file: file path of the stl_file
  :type stl_file: str
  :param file_units: units of the stl_file (``mm`` or ``inch``)
  :type file_units: str
  :param toolID: tool ID
  :type toolID: str
  :return: response dictionary
  :rtype: dict
  """
  props = server.properties
  
  # Check user input
  if stl_file is None: # use demo STL file
    stl_file = os.sep.join(os.path.abspath(__file__).split(os.sep)[:-2] + ['files', 'stls', 'Cylinder.stl'])    
  
  if not toolID:
    toolID = props['toolID']    
    
  if not os.path.isfile(stl_file):
    raise IOError("STL file `{}` does not exist".format(stl_file))
  if not file_units.lower() in ['mm', 'inch']:
    raise ValueError("Unsupported unit '{}' (should be one of Mm or inch)".format(file_units))
  file_units = dict(mm='Mm', inch='inch')[file_units.lower()]
  
  # Create the form with simple fields and add the file
  form = httputils.MultiPartForm()
  form.add_field('fileUnits', file_units)
  form.add_stl('file', stl_file)
  
  # Build the request
  url = 'https://{server_name}/web-api/tool/{toolID}/model'.format(server_name=props['server_name'], toolID=toolID)
  request = urllib_request.Request(url)
  data = form.tobytes()
  request.add_header('Content-type', form.get_content_type())
  request.add_header('Content-length', len(data)) # probably unnecessary
  
  # Send the request and catch the response
  response = urllib_request.urlopen(request, data)
  output = json.loads(response.read().decode())
  return output
  
  
def print_lab_connection(stl_file=None, toolID=None, material=None, finish=None,
                         open_browser=False, currency='EUR', scale=100.0):
  """
  Upload a 3D model to the i.materialise server and get a link to the 3D Print Lab in return
  
  :param scale: scale of the model in percentage (defaults to 100%)
  :type scale: float
  """
  props = server.properties
  materialID, finishID = info.get_material_and_finish_ID(material, finish)
  return_link = None
  
  # Check user input
  if stl_file is None: # use demo STL file
    stl_file = os.sep.join(os.path.abspath(__file__).split(os.sep)[:-2] + ['files', 'stls', 'Cylinder.stl'])    
  
  if not toolID:
    toolID = props['toolID']    
  
  logger.debug("Uploading STL file {}".format(stl_file))
  if not os.path.isfile(stl_file):
    raise IOError("STL file `{}` does not exist".format(stl_file))
  
  # Get server configuration  
  new_location = []
  
  # Create the form with simple fields
  form = httputils.MultiPartForm()
  form.add_field('useAjax', '"False"')
  form.add_field('forceEmbedding', '"False"')
  form.add_field('plugin', toolID)
  form.add_field('materialID', materialID)
  form.add_field('finishID', finishID)
  form.add_field('currency', currency)
  form.add_field('scale', str(scale))
  form.add_stl('file', stl_file)

  # Local class to deal with the redirecting
  class RedirectFilter(urllib_request.HTTPRedirectHandler):    
    def redirect_request(self, req, fp, code, msg, hdrs, newurl):
        new_location.append(newurl)
        return None # do not redirect, HTTPError will be raised

  # Build the request
  opener = urllib_request.build_opener(RedirectFilter)
  urllib_request.install_opener(opener)
  url = 'https://{server_name}/upload'.format(server_name=props['server_name'])
  logger.debug("3D Print Lab upload url: {}".format(url))
  request = urllib_request.Request(url)
  body = form.tobytes()
  request.add_unredirected_header('Content-type', form.get_content_type())
  request.add_unredirected_header('Content-length', len(body))

  # Try to open the URL, we made it such that it will raise an HTTPError when
  # a redirection occurs; but then the new_location contains a url
  try:
    response = urllib_request.urlopen(request, body)
  except urllib_error.HTTPError:
    return_link = new_location[0]
  
  logger.debug("3D Print Lab redirect link: {}".format(return_link))

  # If requested, open the link in a webbrowser
  if return_link and open_browser:
    webbrowser.open(return_link)
  
  return return_link