﻿"""
Pricing APIs

Get information on pricing of a model.

.. autosummary::
  
  pricing_by_parameters
  pricing_by_model
  
"""
from imaterialise.api import server
from imaterialise.api import info
from imaterialise.api import ordering
if server.PY2:
  import urllib2 as urllib_request
else:
  import urllib.request as urllib_request
import json
import logging
import os
import inspect

logger = logging.getLogger("API.PRICING")

def pricing_by_parameters(toolID=None, modelReference='na', material=None,
                          finish=None, quantity=1, xDimMm=12, yDimMm=12, zDimMm=12,
                          volumeCm3=2.0, surfaceCm2=100.0, currency='EUR',
                          shipmentInfo=None):
  """
  Return pricing by parameters.
  
  The response is a nested dictionary with level-1 keys:
    - shipmentCost (dict)
    - currency (str)
    - disclaimer (str)
    - models (list of dict)
  
  Each model dict has keys:
    - finishID and finishingName
    - materialID and materialName
    - toolID
    - modelReference
    - quantity
    - surfaceCm2 and volumeCm3
    - xDimMm, yDimMm, zDimMm    
  
  See `Pricing by parameters API <https://i.materialise.com/api/docs/pricing-by-parameters-api/>`_
  for more information on the response.
  
  The argument ``shipmentInfo`` should be a dict with shipment information,
  see :py:func:`imaterialise.info.get_shipment_info`.
  
  The arguments ``material`` and ``finish`` need to come from the material catalog. If they
  are not given a default material is chosen (which is the first in the materials catalog,
  likely ``Polyamide`` with finish ``Natural white``).
  
  Used information:
    - toolID
    - API key code
      
  Errors:
    - when the material or finish name are invalid, a ``ValueError`` is raised
    - when the server cannot be reached, a ``URLError`` is raised
      
  Example usage::

    >>> response = pricing_by_parameters(material='Silver', finish='Gloss',
                            quantity=2, xDimMm=10, yDimMm=10, zDimMm=10,
                            volumeCm3=1, surfaceCm2=6)
    >>> print(response['models'][0]['totalPrice'])
    70.52
    
  """
  props = server.properties
  materials = info.materials_catalog()
  
  if toolID is None:
    toolID = props['toolID']
    
  if shipmentInfo is None:
    shipmentInfo = info.get_shipment_info()
  elif not shipmentInfo:
    shipmentInfo = dict()
  
  materialID, finishID = info.get_material_and_finish_ID(material, finish)
      
  url = "https://{server_name}/web-api/pricing".format(**props)
  logger.debug("Quering pricing by parameters via {}".format(url))
  
  req = urllib_request.Request(url)
  req.add_header('Content-Type', 'application/json; charset=utf-8')
  req.add_header('ApiCode', props['APIkey'])

  # Get and encode the requested message  
  data = dict(models=[dict(toolID=toolID, modelReference=modelReference,
                           materialID=materialID, finishID=finishID, quantity=str(quantity),
                           xDimMm=str(xDimMm), yDimMm=str(yDimMm), zDimMm=str(zDimMm),
                           volumeCm3=str(volumeCm3), surfaceCm2=str(surfaceCm2))],
                      shipmentInfo=shipmentInfo,
                      currency=currency)
  
  packet = json.dumps(data).encode('utf-8')
  req.add_header('Content-Length', len(packet))
  
  # Send the request to the server and report the response
  logger.debug("Sending: {}".format(str(packet, encoding='utf-8')))

  response = urllib_request.urlopen(req, packet)
  contents = response.read().decode('utf-8')

  logger.debug("Response: {}".format(contents))
  output = json.loads(contents)
  
  return output


def pricing_by_model(stl_file=None, file_units='mm', toolID=None,
                     modelReference=None, material=None,
                     finish=None, quantity=None, xDimMm=None, yDimMm=None, zDimMm=None,
                     volumeCm3=None, surfaceCm2=None,
                     currency='EUR', shipmentInfo=None):
  """
  Get the pricing of a model via the STL file.
  
  This is a wrapper around :py:func:`ordering.upload_model <imaterialise.ordering.upload_model>`
  and :py:func:`pricing.pricing_by_parameters <pricing_by_parameters>`.
  
  Any of the parameters passed to this function will override the answer
  given by the server, if it exists. 
  
  If ``stl_file`` is not given, a demo file will be used::
    
    >>> demo_file = info.get_demo_stl()
    >>> model_params, price = pricing_by_model(demo_file)
    >>> print(price['models'][0]['totalPrice'])
    12.75
    >>> print(model_params['surfaceCm2'])
    25.1186
  
  :return: model parameters and pricing info
  :rtype: (dict, dict)
  """
  props = server.properties
  materials = info.materials_catalog()
  
  if toolID is None:
    toolID = props['toolID']
    
  if shipmentInfo is None:
    shipmentInfo = info.get_shipment_info()
  elif not shipmentInfo:
    shipmentInfo = dict()
    
  # upload the STL file and retrieve the parameters  
  params = ordering.upload_model(stl_file, file_units, toolID)  
  
  # prepare the received parameters to pass on to the pricing_by_parameters function
  pricing_parameters = inspect.getargspec(pricing_by_parameters).args
  params_for_pricing = {key:params[key] for key in pricing_parameters if key in params}
  
  # override any information given locally
  for key in pricing_parameters:
    if key in locals() and locals()[key] is not None:
      params_for_pricing[key] = locals()[key]
  
  # And request the price!
  price = pricing_by_parameters(**params_for_pricing)
  
  return params, price