"""
Some functions and variables to deal with unit conversions in Blender.

Allows pretty print of numbers and their units.

Restricted to units of length, area or volume.
"""
import re
import collections

# We assume mm by default, so a factor 1000.0 is necessary in other systems
# (Blender assumes m)
unit_system_scale_factor = dict(NONE=1.0, METRIC=1000.0, IMPERIAL=1000.0)
unit_system_units = dict(NONE='mm', METRIC='m', IMPERIAL="ft")

# scaling factors for unit prefixes
_scalings ={'y':       1e-24, # yocto
            'z':       1e-21, # zepto
            'a':       1e-18, # atto
            'f':       1e-15, # femto
            'p':       1e-12, # pico
            'n':       1e-09, # nano
            'mu':      1e-06, # micro
            'u':       1e-06, # micro
            'm':       1e-03, # milli
            'c':       1e-02, # centi
            'd':       1e-01, # deci
            'da':      1e+01, # deka
            'h':       1e+02, # hecto
            'k':       1e+03, # kilo
            'M':       1e+06, # mega
            'G':       1e+09, # giga
            'T':       1e+12, # tera
            'P':       1e+15, # peta
            'E':       1e+18, # exa
            'Z':       1e+21, # zetta
            'Y':       1e+24  # yotta
            }

# all factors taken from Wikipedia
_factors = collections.OrderedDict([
# DISTANCE
           ('m',     (  1e+00,       'm','length','SI','meter')), # meter
           ('th',    (0.0000254,     'm','length','imperial','thou (international)')),
           ('in',    (0.0254,        'm','length','imperial','inch (international)')),
           ('ft',    (0.3048,        'm','length','imperial','foot (international)')),
           ('yd',    (0.9144,        'm','length','imperial','yard (international)')),
           ('ch',    (20.1168,       'm','length','imperial','chain (international)')),
           ('fur',   (201.168,       'm','length','imperial','furlong (international)')),
           ('mi',    (1609.344,      'm','length','imperial','mile (international)')),
# AREA
           ('a',     (100.,          'm2','area','SI','are')),
           ('perch', (25.29285264,   'm2','area','imperial','square rod, (square) pole, (square) perch')),
           ('rood',  (1011.7141056,  'm2','area','imperial','rood')),
           ('acre',  (4046.8564224,  'm2','area','imperial','acre (international)')),           
# VOLUME
           ('l',     (  1e-3,        'm3','volume','SI','liter')),
           ('floz',  (28.4130625e-6, 'm3','volume','imperial','fluid ounce')),
           ('gi',    (142.0653125e-6,'m3','volume','imperial','gill (international)')),
           ('pt',    (568.26125e-6,  'm3','volume','imperial','pint (international)')),
           ('qt',    (1136.5225e-6,  'm3','volume','imperial','quart (international)')),
           ('gal',   (4.54609e-3,    'm3','volume','imperial','gallon (international)')),
           ])

_names = {'m1':'length',
          'm2':'area',
          'm3':'volume'}

# Subselection of all units that are considered "readable",
# it's difficult to image dam2, beter then to give in m2, ha or km2
_readable_units = {'length': {'SI':['nm', 'um', 'mm', 'cm', 'm', 'km'],
                              'imperial':['th', 'in', 'ft', 'yd', 'mi']},
                   'area': {'SI':['mm2', 'cm2', 'm2', 'ha', 'km2'],
                            'imperial':['perch', 'acre']},
                   'volume': {'SI':['mm3', 'cm3', 'dm3', 'm3', 'km3'],
                              'imperial':['floz', 'pt', 'gal', 'thou3', 'in3', 'ft3']}
                  }

def convert(_from, _to, value):
  """Convert one unit to another.
  
  Example usage:
  
  >>> value_in_ft = convert('cm', 'ft', 0.2)
  >>> print(value_in_ft)
  0.00656167979003  
  >>> value_in_l = convert('cm3', 'l', 10.)
  >>> print(value_in_l)
  0.01
  
  :param _from: unit to convert from
  :type _from: str
  :param _to: unit to convert to
  :type _to: str
  :param value: value to convert
  :type value: float
  :return: converted value
  :rtype: float
  """
  fac_from, uni_from = breakdown(_from)
  fac_to, uni_to = breakdown(_to)  
  if uni_from!=uni_to:
    raise ValueError("Cannot convert {} to {}".format(uni_from, uni_to))  
      
  return fac_from / fac_to * value    


def breakdown(unit):
  """Decompose a unit into SI base units containing powers.
  
  Example usage:
  
  >>> print(breakdown('in'))
  (0.0254, 'm1')
  
  Or more adventurous:
  
  >>> print(breakdown('10e+03ft3')
  (28316846.592, 'm3')
  
  :param unit: unit's name
  :type unit: str
  :return: 2-tuple factor, unit's base name
  :rtype: (float,str)
  """
  # break down in basic units
  units = unit.split()
  total_factor = 1.
  total_units = []
  total_power = []
  for unit in units:
    factor, basis, power, _ = components(unit)
    total_factor = total_factor * factor**power
    basis = basis.split()
    for base in basis:
      factor_, basis_, power_, _ = components(base)
      try:
        total_factor = total_factor * factor_
      except TypeError:
        pass
      if basis_ in total_units:
        index = total_units.index(basis_)
        total_power[index] += power_ * power
      else:
        total_units.append(basis_)
        total_power.append(power_ * power)  
  # make sure to return a sorted version
  total_units = sorted(['{}{}'.format(i,j) for i,j in zip(total_units,total_power) if j!=0])
  return total_factor, " ".join(total_units)


def components(unit):
  """Decompose a unit into: factor, SI base units, power.
  
  Example usage (compare to ``breakdown``):
  
  >>> print(components('cm'))
  (0.01, 'm', 1, 'SI')
  >>> print(components('10e+03ft3'))
  (304.8, 'm', 3, 'imperial')
  
  :param unit: unit name
  :type unit: str
  :return: 3-tuple with factor, SI base unit, power, unit system
  :rtype: (float, str, int, str)
  """
  unit_system = 'SI'
  factor = 1.0
  # check if there is a prefactor of the form '10e-14' and expand if necessary
  m = re.search('\d\d[[eE][-+]\d\d', unit[:6])
  if m is not None:
    power = float(m.group(0)[3]+'1') * float(m.group(0)[4:6])
    factor *= float(m.group(0)[:2])**power
    unit = unit[6:]
  if not unit[-1].isdigit():
      unit += '1'
  # decompose unit in base name and power
  m = re.search(r'(\d*)(.+?)(-{0,1}[\d\.]+)', unit)
  if m is not None:
    factor_, basis, power = m.group(1), m.group(2), m.group(3)
    # try to make power an integer, otherwise make it a float
    power = float(power) if '.' in power else int(power)
    if factor_:
        factor *= float(factor_)
  else:
    factor, basis, power = 1.0, unit, 1
  # decompose the base name (which can be a composition of a prefix
  #   (e.g., 'mu') and a unit name (e.g. 'm')) into prefix and unit name
  # check if basis is part of _factors dictionary. If not, find the combination
  # of _scalings and basis which is inside the dictionary!
  if basis in _factors:
    unit_system = _factors[basis][3]
  for scale in _scalings:
    scale_unit, base_unit = basis[:len(scale)], basis[len(scale):]
    if scale_unit==scale and base_unit in _factors and not basis in _factors:
      factor *= _scalings[scale]
      basis = base_unit
      unit_system = _factors[base_unit][3]
      break
  # if we didn't find any scalings, check if the 'raw' unit is already a base
  # unit
  else:
    if not basis in _factors:
      raise ValueError('Unknown unit {}'.format(basis))      
  factor *= _factors[basis][0]
  basis = _factors[basis][1]
    
  return factor, basis, power, unit_system


def get_unit_type(unit):
  """Return type of unit (length, area, ...)
  
  :param unit: name of the unit to get the type of
  :type unit: str
  :return: type of the unit (length, area, ...)
  :rtype: str
  """
  comps = breakdown(unit)[1]
  if comps in _names:
    return _names[comps]
  for i in _factors:
    test_against = breakdown(i)[1]
    if test_against==comps or unit==test_against:
      return _factors[i][2]
  return unit


def convert_to_readable(value, unit, allow=None, low=1.0, high=10000.0):
  """Return the unit in which the values is smallest but above 1"""
  if allow is None:
    _, _, _, system = components(unit)
    allow = _readable_units[get_unit_type(unit)][system]
  use_unit = unit
  smallest_value = 1e300
  # Run over all allowed units, and check if the converted value lies between
  # 1 and 10000. We're looking for the smallest such number.
  for check_unit in allow:
    cvalue = convert(unit, check_unit, abs(value))
    if ((low-cvalue)<1e-8) and ((cvalue-high)<1e-8) and (cvalue<smallest_value):
      smallest_value = cvalue
      use_unit = check_unit
  # Finally also convert the value to this unit
  cvalue = convert(unit, use_unit, value)
  return cvalue, use_unit
  