﻿"""
Main script for imaterialise addon.

This script is devided in four parts:
    1. UI panel definitions and callback functions
    2. Operators
    3. Utilities
    4. Obligatory Blender interface definitions
"""

bl_info = {
        "name": "i.materialise Tools", 
        "description": "Price quotation from i.materialise 3D Printing service.",
        "author": "Pieter Degroote, Nils Faber",
        "version": (0, 10),
        "location": "3D View > Toolbox",
        "category": "Object"
}
# Standard library imports
import os
import sys
import webbrowser
import tempfile
import shutil
import zipfile
import urllib
import urllib.request
import math
import logging
# Blender imports
import bpy
import bmesh
import mathutils
# Imaterialise subpackage imports (comes with addon)
from imaterialise.api import ordering
from imaterialise.api import pricing
from imaterialise.api import info
from imaterialise.api import server
from imaterialise.blender import report
from imaterialise.blender import units

_debug = False

unit_system_scale_factor = dict(NONE=1.0, METRIC=1000.0, IMPERIAL=1000.0)
unit_system_units = dict(NONE='mm', METRIC='m', IMPERIAL="ft")
__package_url__ = "https://i.materialise.com/downloads/plugins/imaterialise-blender-addon.zip"
#__package_url__ = r"file:C:\imaterialise-blender-addon-0.6.zip"
logger = logging.getLogger("iMat")

################
# UI Elements
################
                                 
class IMatPanel(bpy.types.Panel):
    """Main Addon UI Panel"""
    bl_space_type = "VIEW_3D"
    bl_region_type = "TOOLS"
    bl_context = "objectmode"
    bl_label = "i.materialise"
    bl_category = "i.materialise"
    
    #@classmethod
    #def poll(cls, context):
    #  return (context.mode in ['OBJECT', 'EDIT_MESH'])
    
    def draw(self, context):
        layout = self.layout
        scn = context.scene
        settings = scn.i_materialise
        
        # Price Quote operator
        box = layout.box()
        box.row().prop(settings, 'matenum')
        box.row().prop(settings, 'colenum')
        box.row().prop(settings, 'quantity', toggle=True)        
        box.row().prop(settings, 'currency', expand=True)
        box.row().operator("mesh.imaterialise_get_price_quote", text="Get price quote (*)")
        
        # 3D Print lab connection
        col = box.row()
        col.operator("mesh.imaterialise_upload", text="Upload 3D model and continue to i.materialise")
                
        # Information Panel & disclaimer
        info = report.info("mesh.imaterialise_get_price_quote")
        box = layout.box()
        for (text, data) in info:
          box.label(text)
        disclaimer = layout.column_flow()
        disclaimer.label("(*) Prices are listed exluding taxes and shipment costs.")        
        layout.separator()     
        
        layout.separator()
        # Updates
        col = layout.column(align=True)        
        col.operator("mesh.imaterialise_check_updates", text="Check for updates")


def get_materials(self, context):
  """Build the drop-down menu listing the different materials"""
  materials = info.materials_catalog()['materials']
  items = [('matenum_{}_{}'.format(i, mat['name']), mat['name'], mat['name'])\
                       for i, mat in enumerate(materials)]
  return items
  
  
def get_finish(self, context):
  """Build the drop-down menu listing the different finishes for a material."""
  selected_material = "_".join(context.scene.i_materialise.matenum.split('_')[2:])
  material_info = info.get_material_info(selected_material)
  items = [('colenum_{}_{}'.format(i, mat['name']), mat['name'], mat['name'])\
                       for i, mat in enumerate(material_info['finishes'])]
  return items
  
  
################
# Operators
################
        
class IMatGetPriceQuote(bpy.types.Operator):
  """Get price quote for the current active object"""
  bl_idname = "mesh.imaterialise_get_price_quote"
  bl_label = "Get price quote for the current active object"

  def execute(self, context):
    """
    Contact the server to get a price quote for the currently selected model.
    
    Scaling handling:
        - if no units are set, mm are assumed
        - if metric, m are assumed
        - if imperial, ft are assumed
    
    On top of that, the global scale factor is also interpreted; if metric
    but scale=0.001, all measurements are in mm.
    
    Error handling:
        - if something goes wrong while contacting the server (service
          not available, dead link, server down etc...) a message is shown
          in the information panel of the addon
        - if the server is up and can be contacted but the parameters are
          wrong for the model (e.g. wrong material etc...), the message is
          also shown in the information panel of the addon.
    """
    logger.debug("Executing GetPriceQuote")
    # Global settings of units and scales
    obj = bpy.context.active_object
    unit = context.scene.unit_settings
    scale_factor = unit_system_scale_factor[unit.system]
    global_scale = unit.scale_length if (unit.system != 'NONE') else 1.0
    unit_name = unit_system_units[unit.system]    
    
    # Get info on material and finishing
    selected_material = "_".join(context.scene.i_materialise.matenum.split('_')[2:])
    selected_finish = "_".join(context.scene.i_materialise.colenum.split('_')[2:])        
    quantity = context.scene.i_materialise.quantity          
    currency = context.scene.i_materialise.currency.split('_')[-1]
    logger.debug("Selected material={}, finish={}".format(selected_material, selected_finish))
    logger.debug("Selected quantity={}, currency={}".format(quantity, currency))
    
    # Check if object is a MESH
    if obj.type != 'MESH':
      report_info = [("Error: Please convert object from {} to MESH (Alt C)".format(obj.type), None)]
      report.update(*report_info, label="mesh.imaterialise_get_price_quote")
      return {'FINISHED'}

    # Retrieve the pricing based on the current model. If it fails, exit silently
    try:
      volume = calculate_volume(obj)*scale_factor**3*global_scale**3 / 1000.0
      area = calculate_area(obj)*scale_factor**2*global_scale**2 / 100.0
      response = pricing.pricing_by_parameters(
          xDimMm=obj.dimensions.x*scale_factor*global_scale,
          yDimMm=obj.dimensions.y*scale_factor*global_scale,
          zDimMm=obj.dimensions.z*scale_factor*global_scale,
          volumeCm3=volume, surfaceCm2=area,
          material=selected_material, finish=selected_finish,
          quantity=quantity, currency=currency
          )
    except urllib.error.URLError as e:
      str_reason = str(e.reason)
      if ('getaddrinfo' in str_reason):
        report_info = [("Error: please check internet connection", None)]
      elif ('10061' in str_reason):
        report_info = [("Error: i.materialise server is not responding, please try again later", None)]
      else:
        report_info = [("URLError: {}".format(str(e)), None)]        
      report.update(*report_info, label="mesh.imaterialise_get_price_quote")
      return {'FINISHED'}
    except Exception as e:
      report_info = [("Error: {}".format(str(e)), None)]
      report.update(*report_info, label="mesh.imaterialise_get_price_quote")
      return {'FINISHED'}
    
    # Construct the response string for the information panel        
    str_template = [
            "Model '{name}': {model[totalPrice]:6.2f} {response[currency]}(*)",
            "Volume: {model[volumeCm3]:.4g} {model[unit_volume]}",
            "Surface: {model[surfaceCm2]:.4g} {model[unit_area]}",
            "Dimension: ({model[xDimMm]:.4g}, {model[yDimMm]:.4g}, {model[zDimMm]:.4g}) {model[unit_length]}",
            "Material: {model[materialName]}",
            "Color/Finish: {model[finishingName]}",
            "Quantity: {model[quantity]}",
            ]
    report_info = []         
    for i, model in enumerate(response['models']):
      # Check if it was possible to calculate pricing:
      if 'pricingError' in model:
        report_info.append(("Model {name}: {error[message]}".format(name=obj.name,
                                    error=model['pricingError']), None))
        continue
      # Now we try to convert the volume, surface and length to "readable" units,
      # i.e. instead of 10000mm we take 100 m and so on (but we stay in the same unit system):
      model['volumeCm3'] = units.convert('cm3', unit_name+'3', model['volumeCm3'])
      model['surfaceCm2'] = units.convert('cm2', unit_name+'2', model['surfaceCm2'])
      model['xDimMm'] = units.convert('mm', unit_name, model['xDimMm'])
      model['volumeCm3'], model['unit_volume'] = units.convert_to_readable(model['volumeCm3'], unit_name+'3')
      model['surfaceCm2'], model['unit_area'] = units.convert_to_readable(model['surfaceCm2'], unit_name+'2')
      model['xDimMm'], model['unit_length'] = units.convert_to_readable(model['xDimMm'], unit_name)
      model['yDimMm'] = units.convert('mm', model['unit_length'], model['yDimMm'])
      model['zDimMm'] = units.convert('mm', model['unit_length'], model['zDimMm'])
      # Finally construct the strings for the report
      for istr_tmpl in str_template:
        report_info.append((istr_tmpl.format(name=obj.name, response=response, model=model), obj))
    report.update(*report_info, label="mesh.imaterialise_get_price_quote")        
    return {'FINISHED'}


class IMatUpload(bpy.types.Operator):
  """Upload model to 3D print lab"""
  bl_idname = "mesh.imaterialise_upload"
  bl_label = "Upload model to imaterialise"

  def execute(self, context):
    # Global info on units and scaling
    unit = context.scene.unit_settings
    global_scale = unit.scale_length if (unit.system != 'NONE') else 1.0
    global_scale *= unit_system_scale_factor[unit.system]
    currency = context.scene.i_materialise.currency.split('_')[-1]
    # Get info on material and finishing
    material = "_".join(context.scene.i_materialise.matenum.split('_')[2:])
    finish = "_".join(context.scene.i_materialise.colenum.split('_')[2:])
    # Export the currently selected object to a temporary STL file, which can
    # then be uploaded. Apply all mesh modifiers and scalings before uploading.
    # The STL file is binary and will be deleted after this function exits
    try:
      temp_dir = tempfile.mkdtemp()
      logger.debug("Created temporary directory for STL export: {}".format(temp_dir))
      temp_file = os.path.join(temp_dir, "_".join(bpy.context.active_object.name.split()) +'.stl')
      ret = bpy.ops.export_mesh.stl(
                context.copy(),
                filepath=temp_file,
                ascii=False,
                use_mesh_modifiers=True,
                global_scale=global_scale,
                )
      if 'FINISHED' in ret:
        # Simple upload
        imat_url = ordering.print_lab_connection(temp_file, material=material,
                       finish=finish, open_browser=True, currency=currency)
    except urllib.error.URLError as e:
      str_reason = str(e.reason)
      if ('getaddrinfo' in str_reason):
        report_info = [("Error: please check internet connection", None)]
      elif ('10061' in str_reason):
        report_info = [("Error: i.materialise server is not responding, please try again later", None)]
      else:
        report_info = [("URLError: {}".format(str(e)), None)]        
      report.update(*report_info, label="mesh.imaterialise_get_price_quote")
      return {'FINISHED'}
    except Exception as e:
      report_info = [("Error: {}".format(str(e)), None)]
      report.update(*report_info, label="mesh.imaterialise_get_price_quote")
      return {'FINISHED'}
    finally:
      # Remove temporary directory and everything in it
      shutil.rmtree(temp_dir)
      logger.debug("Removed temporary directory for STL export: {}".format(temp_dir))
    return {'FINISHED'}

      
class IMatCheckUpdates(bpy.types.Operator):
  """Check for updates of the material catalog and the addon itself"""
  
  bl_idname = "mesh.imaterialise_check_updates"
  bl_label = "Check for updates"
  
  # Updating the package means simply extracting the online zip archive to the
  # current location: existing files are overwritten, new files are automatically
  # added but obsolete files or not removed (though they will not be used anymore
  # either so they can be manually removed).
  #
  # Updating the materials catalog means that the material catalog file in the
  # imaterialise subpackage of the addon will be replaced with a new version
  # extracted from the server (it's a json dump of the server response).
  
  # The add-on update will only have effect on the next restart of Blender.
  
  # The material catalog update has immediate effect.
  
  def execute(self, context):
    # Collect info on addon update and material catalog in one condense message
    # to show in the information panel
    message = "iMat addon:"
    # First update the whole package.
    try:
      # Download the zip file from the server
      filename, headers = urllib.request.urlretrieve(__package_url__)
      # First check for version number: read this file (i.e. imaterialise_addon.py)
      # from the retrieved archive and look for the version number. It's
      # somewhere in the bl_info dictionary which is located before all the
      # imports
      version = None
      with zipfile.ZipFile(filename) as myzip:
        with myzip.open('imaterialise_addon.py','r') as myfile:
          contents = myfile.read().decode('utf-8')
          for line in contents.split('\n'):
            line = line.strip().split()
            if not line:
              continue
            elif 'import' in line[0]:
              break
            elif '"version"' in line[0] or "'version'" in line[0]:
              line = "".join(line)
              version = line.split('(')[1].split(')')[0].split(',')
              version = tuple([int(v) for v in version])
              break
      # If the version number is found, check if it is smaller or equal than
      # the current number.
      if version is not None:
        # If the package is up-to-date, don't do anything
        if version <= bl_info['version']:
          message += " already up-to-date (v{})".format(".".join([str(v) for v in version]))
        # Else, extract all the files to the directory of the current file
        # (i.e. imaterialise_addon.py). This should be writable in all OS.
        else:
          with zipfile.ZipFile(filename) as myzip:
            myzip.extractall(path=os.path.dirname(os.path.abspath(__file__)))
          message += " updated to v{}".format(".".join([str(v) for v in version]))
      # If for some reason, the version number could not be extracted from the
      # downloaded file, report to the info panel
      else:
        message += " update failed"
    # Perhaps we have no internet connection, perhaps the online server is down,
    # perhaps the world is on fire... then report the fail to the user.
    except Exception as e:
      message += " update failed (cannot connect to server)"
    finally:   
      urllib.request.urlcleanup()
    
    # Second, update the materials specifically: the new catalog will replace
    # the old one no matter what, but we'll check for equality anyway to report
    # to the user if anything changed.
    old_materials = report.get_materials()
    try:
      output = info.materials_catalog(force_update=True)
      report.update_materials(output)
      new_materials = report.get_materials()
      if old_materials!=new_materials:
        message += "; iMat catalog: updated"
      else:
        message += "; iMat catalog: up-to-date"
    except urllib.error.URLError as e:
      str_reason = str(e.reason)
      if ('getaddrinfo' in str_reason):
        message += "; iMat catalog: failed (no connection)"
      elif ('10061' in str_reason):
        message += "; iMat catalog: failed (server down)" 
      else:
        message += "; iMat catalog: URLError: {}".format(str(e))      
    except Exception as e:
      message += "; iMat catalog: Error: {}".format(str(e))
      
    self.report({'INFO'}, message)
    return {'FINISHED'}      


######################
# Utilities
######################
  
def bmesh_copy_from_object(obj, transform=True, triangulate=True, apply_modifiers=False):
  """Returns a transformed, triangulated copy of the mesh
  
  Taken from the 3d print utility
  """
  # Handle curves by converting them to meshes:
  if not obj.type == 'MESH':
    bpy.ops.object.convert(target='MESH')
  
  if apply_modifiers and obj.modifiers:
    me = obj.to_mesh(bpy.context.scene, True, 'PREVIEW', calc_tessface=False)
    bm = bmesh.new()
    bm.from_mesh(me)
    bpy.data.meshes.remove(me)
  else:
    me = obj.data
    if obj.mode == 'EDIT':
      bm_orig = bmesh.from_edit_mesh(me)
      bm = bm_orig.copy()
    else:
      bm = bmesh.new()
      bm.from_mesh(me)

  if transform:
    bm.transform(obj.matrix_world)

  if triangulate:
    bmesh.ops.triangulate(bm, faces=bm.faces)

  return bm  
  
  
def calculate_area(obj):
  """Area is computed after applying all modifiers to the object"""
  bm = bmesh_copy_from_object(obj, apply_modifiers=True)
  area = sum(f.calc_area() for f in bm.faces)
  bm.free()  
  return area
  
def calculate_volume(obj):
  """Volume is computed after applying all modifiers to the object"""
  bm = bmesh_copy_from_object(obj, apply_modifiers=True)
  vol = bm.calc_volume(signed=True)
  bm.free()  
  return vol
  

######################
# Blender interface
######################
  
class IMaterialiseSettings(bpy.types.PropertyGroup):
    """Properties that can be accessed in the operators"""
    quantity = bpy.props.IntProperty(
                 name="Quantity", 
                 description="Amount of 3D prints",
                 min=1)
    matenum = bpy.props.EnumProperty(items = get_materials, name = "Material", description="Base material")
    colenum = bpy.props.EnumProperty(items = get_finish, name = "Color/Finish", description="Material color/finish")
    currency = bpy.props.EnumProperty(items = [('currency_0_EUR', 'EUR', 'Euro'),
                                                ('currency_1_USD', 'USD', 'US Dollar'),
                                                ('currency_2_JPY', 'JPY', 'Japanese Yen'),
                                                ('currency_3_GBP', 'GBP', 'British Pound'),],
                                      default='currency_0_EUR', name = "Currency",
                                      description="Currency for price quote")
    set_lighting = bpy.props.BoolProperty(name="Set lighting",
        default=True, description="Setup/update lighting in the scene")  
    set_material = bpy.props.BoolProperty(name="Set material",
        default=True, description="Set object's material node properties")
    set_renderer = bpy.props.BoolProperty(name="Set renderer",
        default=True, description="Adjust renderer settings for improved viewing")
    set_viewport = bpy.props.BoolProperty(name="Set viewport",
        default=True, description="Switch viewport to RENDERED mode")

        
class IMatPanelObject(IMatPanel):
    #bl_category = "3D Printing"
    bl_idname = "MESH_PT_imat_object"
    bl_context = "objectmode"

class IMatPanelMesh(IMatPanel):
    #bl_category = "3D Printing"
    bl_idname = "MESH_PT_imat_mesh"
    bl_context = "mesh_edit"        
        
        
classes = (
    IMatPanelObject,
    IMatPanelMesh,
    IMatGetPriceQuote,
    IMatUpload,
    IMatCheckUpdates,
    IMaterialiseSettings
    )

def register():
    for c in classes:
        bpy.utils.register_class(c)

    bpy.types.Scene.i_materialise = bpy.props.PointerProperty(type=IMaterialiseSettings)

    # Provide logging capabilities in debug mode
    if _debug:
        print("Setup debugging in {}".format(os.path.join(os.path.dirname(os.path.abspath(__file__)))))
        format  = '%(asctime)s %(name)-12s %(levelname)-7s %(message)s'
        datefmt = '%a, %d %b %Y %H:%M'
        formatter = logging.Formatter(fmt=format,datefmt=datefmt)
        filename = os.path.join(os.path.dirname(os.path.abspath(__file__)), "imaterialise_debug_info.log")
        logging.basicConfig(level="DEBUG",
                            format=format,datefmt=datefmt,
                            filename=filename,
                            filemode='w')
        #fh = logging.FileHandler(filename)
        #logging.getLogger("").addHandler(fh)
        logger.debug("Registered Add-on")
    
def unregister():
    for c in classes:
        bpy.utils.unregister_class(c)

    del bpy.types.Scene.i_materialise

if __name__ == "__main__":
    register()    